"""APIs."""

from datetime import datetime
from functools import cached_property

import humanize
import requests
import typer
from pydantic import BaseModel
from pytz import UTC
from slices_bi_client.api.flavors_api import FlavorsApi as BIFlavorsApi
from slices_bi_client.api.images_api import ImagesApi as BIImagesApi
from slices_bi_client.api.resources_api import ResourcesApi as BIResourcesApi
from slices_bi_client.api.tasks_api import TasksApi as BITasksApi
from slices_bi_client.api_client import ApiClient as BIApiClient
from slices_bi_client.configuration import Configuration as BIConfiguration
from slices_expauth_client.api.experiments_api import ExperimentsApi
from slices_expauth_client.api_client import ApiClient as ExperimentApiClient
from slices_expauth_client.configuration import Configuration as ExperimentApiConfiguration
from slices_expauth_client.exceptions import ForbiddenException as ExperimentForbiddenException
from slices_expauth_client.models.experiment import Experiment
from slices_expauth_client.models.experiment_create_v12 import ExperimentCreateV12
from slices_projauth_client.api.project_api import ProjectApi
from slices_projauth_client.api_client import ApiClient as ProjectApiClient
from slices_projauth_client.configuration import Configuration as ProjectApiConfiguration
from slices_projauth_client.exceptions import ForbiddenException as ProjectForbiddenException
from slices_projauth_client.models.project import Project
from slices_pubkey_client.api.public_keys_api import PublicKeysApi
from slices_pubkey_client.api_client import ApiClient as PubkeyApiClient
from slices_pubkey_client.configuration import Configuration as PubkeyConfiguration

from . import console, err_console
from .config import load_auth_token
from .oauth2.config import load_oauth_config
from .oauth2.token import OAuthToken
from .utils import friendly_datetime


class ApiEntry(BaseModel):
    """Information needed to connect to API."""

    base_url: str
    audience: str


class ApiClientsConfig(BaseModel):
    """Configuration model used to create api client objects."""

    project_authority: ApiEntry
    experiment_authority: ApiEntry
    pubkey: ApiEntry
    orchestrator: ApiEntry
    bi: dict[str, ApiEntry]


class NotLoggedInException(Exception):
    """Not logged in."""


class ApiClients:
    """API Clients."""

    auth_token: OAuthToken | None
    config: ApiClientsConfig

    def __init__(self, config: ApiClientsConfig):
        """Create API clients."""
        self.auth_token = load_auth_token()
        self.oauth_config = load_oauth_config()
        self.config = config

    @cached_property
    def projects_api(self):
        """Projects API."""
        proj_config = ProjectApiConfiguration(
            host=str(self.config.project_authority.base_url),
            access_token=self.auth_token["access_token"] if self.auth_token else None,
        )
        proj_client = ProjectApiClient(proj_config)
        return ProjectApi(proj_client)

    @cached_property
    def pubkey_api(self):
        """Pubkey API."""
        pubkey_config = PubkeyConfiguration(
            host=str(self.config.pubkey.base_url),
            access_token=self.get_access_token_for_audience(self.config.pubkey.audience),
        )
        pubkey_client = PubkeyApiClient(pubkey_config)
        return PublicKeysApi(pubkey_client)

    @cached_property
    def experiments_api(self):
        """Experiments API."""

        exp_config = ExperimentApiConfiguration(
            host=str(self.config.experiment_authority.base_url),
            access_token=self.get_access_token_for_audience(self.config.experiment_authority.audience),
        )
        exp_client = ExperimentApiClient(exp_config)

        return ExperimentsApi(exp_client)

    def get_id_token_for_audience(self, audience: str, project_id: str):
        """Return JWT ID token of logged in user."""
        if not self.auth_token:
            raise NotLoggedInException

        data = {
            "grant_type": "urn:ietf:params:oauth:grant-type:token-exchange",
            "subject_token": self.auth_token["access_token"],
            "subject_token_type": "urn:ietf:params:oauth:token-type:access_token",
            "client_id": self.oauth_config["client_id"],
            "client_secret": self.oauth_config["client_secret"],
            "requested_token_type": "urn:ietf:params:oauth:token-type:id_token",
            "audience": audience,
            "scope": f"{self.auth_token['scope']} project:{project_id}",
        }

        response = requests.post(
            url=self.oauth_config["token_endpoint_url"],
            data=data,
            timeout=10,
        )
        response.raise_for_status()
        return response.json()["access_token"]

    def get_access_token_for_audience(self, audience: str, project_id: str | None = None):
        """Return JWT access token of logged in user for target resource."""
        if not self.auth_token:
            raise NotLoggedInException

        scope = self.auth_token["scope"]
        if project_id:
            scope = f"{scope} project:{project_id}"

        data = {
            "grant_type": "urn:ietf:params:oauth:grant-type:token-exchange",
            "subject_token": self.auth_token["access_token"],
            "subject_token_type": "urn:ietf:params:oauth:token-type:access_token",
            "client_id": self.oauth_config["client_id"],
            "client_secret": self.oauth_config["client_secret"],
            "requested_token_type": "urn:ietf:params:oauth:token-type:access_token",
            "audience": audience,
            "scope": scope,
        }

        response = requests.post(
            url=self.oauth_config["token_endpoint_url"],
            data=data,
            timeout=10,
        )
        response.raise_for_status()
        return response.json()["access_token"]

    def get_experiment(self, experiment_name: str, *, project_id: str, must_be_active=False) -> Experiment | None:
        """Get experiment by name."""
        result_page = self.experiments_api.get_experiments(
            friendly_name=experiment_name,
            project_id=project_id,
            include_archived=not must_be_active,
            include_expired=not must_be_active,
        )
        if len(result_page.experiments) > 0:
            return result_page.experiments[0]
        return None

    def get_experiment_by_id(self, experiment_id: str) -> Experiment:
        """Get Experiment by ID."""
        return self.experiments_api.get_experiment(experiment_id=experiment_id)

    def get_or_create_experiment(self, experiment_name: str, project_id: str, expires_at: datetime) -> Experiment:
        """Get or create experiment by name."""
        experiment = self.get_experiment(
            experiment_name,
            project_id=project_id,
            must_be_active=True,
        )

        if not experiment:
            # Create it

            user_project_id_token = self.get_id_with_project_jwt(self.config.experiment_authority.audience, project_id)

            # First make sure project has not yet expired.
            project = self.projects_api.get_project(project_id)
            if project and project.expires_at and project.expires_at < datetime.now(UTC):
                hum_ago = humanize.precisedelta(
                    project.expires_at - datetime.now(UTC), minimum_unit="hours", format="%d"
                )
                console.print(
                    f"[red]:clock10: Project expired on" f" {friendly_datetime(project.expires_at)} ({hum_ago} ago)."
                )
                raise typer.Exit(code=1) from None

            experiment = self.experiments_api.create_experiment(
                ExperimentCreateV12(
                    friendly_name=experiment_name,
                    description=None,
                    expires_at=expires_at,
                    user_project_id_token=user_project_id_token,
                )
            )

        return experiment

    def get_experiment_jwt(self, experiment_id: str) -> str:
        """Get Experiment JWT."""
        params = self.experiments_api.api_client.param_serialize(
            method="GET",
            resource_path="/apis/exp.slices-ri.eu/v1alpha1/experiments/{experiment_id}",
            path_params={"experiment_id": experiment_id},
            header_params={"Accept": "application/jwt"},
            auth_settings=["JWTBearer"],
        )
        try:
            response = self.experiments_api.api_client.call_api(*params)

            return response.response.data.decode("utf-8")
        except ExperimentForbiddenException as err:
            err_console.print(
                f"[red]You have insufficient permissions to create a token for experiment [i]{experiment_id}[/] . :x:"
            )
            raise typer.Exit(1) from err

    def get_project_jwt(self, project_id: str) -> str:
        """Get Project JWT."""
        params = self.projects_api.api_client.param_serialize(
            method="GET",
            resource_path="/apis/proj.slices-ri.eu/v1alpha1/projects/{project_id}",
            path_params={"project_id": project_id},
            header_params={"Accept": "application/jwt"},
            auth_settings=["jwt"],
        )

        try:
            response = self.projects_api.api_client.call_api(*params)

            return response.response.data.decode("utf-8")
        except ProjectForbiddenException as err:
            err_console.print(
                f"[red]You have insufficient permissions to create a token for project [i]{project_id}[/] . :x:"
            )
            raise typer.Exit(1) from err

    def get_id_with_project_jwt(self, audience: str, project_id: str) -> str:
        """Get ID Token with project JWT."""

        return self.get_id_token_for_audience(audience, project_id)

    def get_project_by_name(self, project_name: str) -> Project | None:
        """Get project by name."""
        projects = self.projects_api.get_projects()
        for project in projects.projects:
            if project_name in (project.id, project.name):
                return project
        return None

    def _create_bi_api_client(
        self,
        bi_site,
        *,
        access_token: str | None = None,
        project_id: str | None = None,
    ) -> BIApiClient:
        if bi_site not in self.config.bi and bi_site != "bi-orchestrator":
            raise ValueError("Invalid BI site")

        api_config = self.config.bi[bi_site] if bi_site != "bi-orchestrator" else self.config.orchestrator

        if not access_token:
            if not project_id:
                raise ValueError("Project ID is required")
            access_token = self.get_access_token_for_audience(api_config.audience, project_id)

        bi_config = BIConfiguration(
            host=str(api_config.base_url),
            access_token=access_token,
        )
        return BIApiClient(bi_config)

    def get_bi_resources_api(self, bi_site, experiment_token) -> BIResourcesApi:
        """Get BI Resources API."""
        bi_client = self._create_bi_api_client(bi_site, access_token=experiment_token)

        return BIResourcesApi(bi_client)

    def get_orchestrator_api(self, experiment_token) -> BIResourcesApi:
        """Get Orchestrator API."""
        bi_client = self._create_bi_api_client("bi-orchestrator", access_token=experiment_token)

        return BIResourcesApi(bi_client)

    def get_bi_tasks_api(self, bi_site, project_id: str) -> BITasksApi:
        """Get BI Tasks API."""
        bi_client = self._create_bi_api_client(bi_site, project_id=project_id)

        return BITasksApi(bi_client)

    def get_bi_flavors_api(self, bi_site, project_id: str) -> BIFlavorsApi:
        """Get BI Flavors API."""
        bi_client = self._create_bi_api_client(bi_site, project_id=project_id)

        return BIFlavorsApi(bi_client)

    def get_bi_images_api(self, bi_site, project_id: str) -> BIImagesApi:
        """Get BI Images API."""
        bi_client = self._create_bi_api_client(bi_site, project_id=project_id)

        return BIImagesApi(bi_client)
